class Field(object):
    __slots__ = (
        'field_type',
        'convert',
        'field_comment',
        'is_nullable',
        'is_primary_key',
        'default',
        'extra',
        'numeric_precision',
        'numeric_scale',
        'character_maximum_length',
        'character_octet_length'
    )

    def __init__(self, field_type, convert, *, field_comment=None,is_nullable=True, is_primary_key=False, default=None, extra=None, numeric_precision=None, numeric_scale=None, character_maximum_length=None, character_octet_length=None):
        self.field_type = field_type
        self.convert = convert
        self.field_comment = field_comment
        self.is_nullable = is_nullable
        self.is_primary_key = is_primary_key
        self.default = default
        self.extra = extra
        self.numeric_precision = numeric_precision
        self.numeric_scale = numeric_scale
        self.character_maximum_length = character_maximum_length
        self.character_octet_length = character_octet_length

    def __repr__(self):
        return f"{self.__class__.__name__}(" \
               + f"'{self.field_type}'," \
               + f"'{self.field_comment}'," \
               + f"is_nullable={self.is_nullable}," \
               + f"is_primary_key={self.is_primary_key}," \
               + f"default={self.default}," \
               + ('extra=None,' if self.extra == '' else f"extra='{self.extra}',") \
               + f"numeric_precision={self.numeric_precision}," \
               + f"numeric_scale={self.numeric_scale}," \
               + f"character_maximum_length={self.character_maximum_length}," \
               + f"character_octet_length={self.character_octet_length}" \
               + f")"


class BaseString(Field):
    def __init__(self, field_type, *args, **kwargs):
        super().__init__(field_type, str, *args, **kwargs)


class BaseInt(Field):
    def __init__(self, field_type, *args, **kwargs):
        super().__init__(field_type, int, *args, **kwargs)


class BaseFloat(Field):
    def __init__(self, field_type, *args, **kwargs):
        super().__init__(field_type, float, *args, **kwargs)


class BaseDateTime(Field):
    def __init__(self, field_type, *args, **kwargs):
        super().__init__(field_type, str, *args, **kwargs)


# 字符
string_list = [
    'char', 'varchar',
    'text', 'tinytext', 'mediumtext', 'longtext',
    'blob', 'tinyblob', 'mediumblob', 'longblob',
    'enum',
]
# 数值
int_list = [
    'int', 'integer', 'tinyint', 'smallint', 'mediumint', 'bigint',
]
# 浮点
float_list = [
    'float', 'double', 'decimal'
]
# 日期时间
datetime_list = [
    'date', 'time', 'year', 'datetime', 'timestamp',
]



import pymysql

from .util import SqlUtil
from .exception import SakuraException
from .log import logger
from .models import Model


class SakuraMysql:
    def __init__(self, *args, **kwargs) -> None:
        self.conn = pymysql.connect(*args, **kwargs)

    def _sql(self, query, args):
        """获取sql语句"""
        cur = self.conn.cursor()
        return cur.mogrify(query, args)

    def insert(self, model, field_value):
        fields, _args = list(field_value.keys()), list(field_value.values())
        _fields = SqlUtil.get_fields(fields)

        _sql = SqlUtil.format_sql(
            [
                'INSERT INTO',
                f'`{model.table}`',
                f'({_fields})',
                f'VALUES ({", ".join(["%s"] * len(_args))})'
            ]
        )
        try:
            cur = self.execute(_sql, _args, True)
            if cur:
                return cur.lastrowid
            else:
                return False
        except Exception as e:
            raise SakuraException(e)

    def update(self, model, field_value, cond=None):
        _field_value, _args1 = SqlUtil.get_field_value(field_value)
        _where, _args2 = SqlUtil.get_where(cond)
        _args = _args1 + _args2
        _sql = SqlUtil.format_sql(
            [
                'UPDATE',
                f'`{model.table}`',
                'SET',
                _field_value,
                _where]
        )
        try:
            if self.execute(_sql, _args, True):
                return True
            else:
                return False
        except Exception as e:
            raise SakuraException(e)

    def delete(self, model, cond=None):
        _where, _args = SqlUtil.get_where(cond)
        _sql = SqlUtil.format_sql(
            [
                'DELETE FROM',
                f'`{model.table}`',
                _where,
            ]
        )
        try:
            cur = self.execute(_sql, _args, True)
            if cur:
                return True
            else:
                return False
        except Exception as e:
            raise SakuraException(e)

    def select(self, model, cond=None, group_by=None, order_by=None, limit=100, fields=None):
        _fields = SqlUtil.get_fields(fields)
        if not fields:
            fields = model.fields
        _where, _args = SqlUtil.get_where(cond)
        _group_by = SqlUtil.get_groupby(group_by)
        _order_by = SqlUtil.get_orderby(order_by)
        _limit = SqlUtil.get_limit(limit)
        _sql = SqlUtil.format_sql(
            [
                'SELECT',
                _fields,
                'FROM',
                f'`{model.table}`',
                _where,
                _group_by,
                _order_by,
                _limit
            ]
        )
        try:
            l = self.execute(_sql, _args)
            models = []
            for i in l:
                models.append(model(**dict(zip(fields, i))))
            return models
        except Exception as e:
            raise SakuraException(e)

    def select_one(self, model, cond=None, group_by=None, order_by=None, fields=None):
        models = self.select(model, cond, group_by, order_by, 1, fields)
        if models:
            return models[0]
        return {}

    def execute(self, query, args=None, commit=False):
        try:
            cur = self.conn.cursor()
            logger.debug('sql:%s', self._sql(query, args))
            cur.execute(query, args)
            if commit:
                self.conn.commit()
                return cur
            return cur.fetchall()
        except Exception as e:
            raise SakuraException(e)

    def execute_many(self, query, args=None):
        try:
            cur = self.conn.cursor()
            logger.debug('sql:%s', self._sql(query, args))
            cur.executemany(query, args)
            self.conn.commit()
            return cur.fetchall()
        except Exception as e:
            raise SakuraException(e)

    def getModel(self, tablename):
        fields = {
            'connection': self
        }
        sql = f'''
        select COLUMN_NAME,COLUMN_COMMENT,DATA_TYPE,IS_NULLABLE,COLUMN_KEY,COLUMN_DEFAULT,EXTRA,NUMERIC_PRECISION,NUMERIC_SCALE,CHARACTER_MAXIMUM_LENGTH,CHARACTER_OCTET_LENGTH from information_schema.COLUMNS
        where TABLE_SCHEMA = '{self.conn.db.decode()}' 
        and TABLE_NAME = '{tablename}'
        order by ORDINAL_POSITION; 
        '''
        for i in self.execute(sql):
            field_name, field_comment,field_type, is_nullable, primary, default, extra, numeric_precision, numeric_scale, character_maximum_length, character_octet_length = i
            Field = SqlUtil.getField(field_type)
            fields[field_name] = Field(
                field_type,
                field_comment=field_comment,
                is_nullable=is_nullable.lower() == 'yes',
                is_primary_key=primary.lower() == 'pri',
                default=default,
                extra=extra,
                numeric_precision=numeric_precision,
                numeric_scale=numeric_scale,
                character_maximum_length=character_maximum_length,
                character_octet_length=character_octet_length
            )
        return type(tablename.title(), (Model,), fields)


def connect(*args, **kwargs):
    return SakuraMysql(*args, **kwargs)
